//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Processed/ProcessedSample.cpp
//! @brief     Implements class ProcessedSample.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Resample/Processed/ProcessedSample.h"
#include "Base/Utils/Assert.h"
#include "Resample/Fresnel/MatrixFresnelMap.h"
#include "Resample/Fresnel/ScalarFresnelMap.h"
#include "Resample/Options/SimulationOptions.h"
#include "Resample/Processed/ParticleRegions.h"
#include "Resample/Processed/ProcessedLayout.h"
#include "Resample/Slice/Slice.h"
#include "Resample/Slice/SliceStack.h"
#include "Resample/Specular/SpecularStrategyBuilder.h"
#include "Sample/Interface/LayerRoughness.h"
#include "Sample/Material/Admixtures.h"
#include "Sample/Material/MaterialBySLDImpl.h"
#include "Sample/Material/MaterialUtils.h"
#include "Sample/Material/RefractiveMaterialImpl.h"
#include "Sample/Multilayer/Layer.h"
#include "Sample/Multilayer/MultiLayer.h"
#include "Sample/Multilayer/MultilayerUtils.h"
#include "Sample/Scattering/ZLimits.h"

namespace {

//! Returns the average of an intensive property.
//! Used for SLD (T=complex) and magnetization (T=vector).

template <class T>
T averageData(const Material& layer_mat, const Admixtures& admixtures,
              std::function<T(const Material&)> average)
{
    const T layer_data = average(layer_mat);
    T averaged_data = layer_data;
    for (const OneAdmixture& admix : admixtures)
        averaged_data += admix.fraction * (average(admix.material) - layer_data);
    return averaged_data;
}

// Tested by Tests/UnitTests/Core/Sample/MaterialTest.cpp

Material createAveragedMaterial(const Material& layer_mat, const Admixtures& admixtures)
{
    const std::string avge_name = layer_mat.getName() + "_avg";

    const kvector_t avge_magnetization = averageData<kvector_t>(
        layer_mat, admixtures, [](const Material& mat) { return mat.magnetization(); });

    // Is there a uniform material type (refractive or SLD)?
    std::vector<const Material*> materials;
    materials.push_back(&layer_mat);
    for (const OneAdmixture& admix : admixtures)
        materials.push_back(&admix.material);

    const MATERIAL_TYPES common_type = MaterialUtils::checkMaterialTypes(materials);
    if (common_type == MATERIAL_TYPES::InvalidMaterialType)
        throw std::runtime_error("Error in createAveragedMaterial(): non-default materials of "
                                 "different types used simultaneously.");

    if (common_type == MATERIAL_TYPES::RefractiveMaterial) {
        // avrData returns (1 - mdc)^2 - 1, where mdc is material data conjugate
        auto avrData = [](const Material& mat) -> complex_t {
            const complex_t mdc = std::conj(mat.materialData());
            return mdc * mdc - 2.0 * mdc;
        };
        const complex_t avr_mat_data = std::conj(
            1.0 - std::sqrt(1.0 + averageData<complex_t>(layer_mat, admixtures, avrData)));
        return HomogeneousMaterial(avge_name, avr_mat_data.real(), avr_mat_data.imag(),
                                   avge_magnetization);
    }
    if (common_type == MATERIAL_TYPES::MaterialBySLD) {
        complex_t (*avrData)(const Material&) = [](const Material& mat) {
            return mat.materialData();
        };
        const complex_t avr_mat_data = averageData<complex_t>(layer_mat, admixtures, avrData);
        return MaterialBySLD(avge_name, avr_mat_data.real(), avr_mat_data.imag(),
                             avge_magnetization);
    }
    ASSERT(0);
}

void checkVolumeFractions(const Admixtures& admixtures)
{
    double total_fraction = 0.0;
    for (const auto& admix : admixtures)
        total_fraction += admix.fraction;
    ASSERT(total_fraction >= 0);
    if (total_fraction > 1)
        throw std::runtime_error("average slices: "
                                 "total volumetric fraction of particles exceeds 1!");
}

SliceStack refineStack(const SliceStack& stack, const std::vector<ProcessedLayout>& layouts)
{
    SliceStack result = stack;
    for (const ProcessedLayout& layout : layouts) {
        for (const auto& entry : layout.regionMap()) {
            const size_t i_slice = entry.first;
            if (i_slice == 0 || i_slice == result.size() - 1)
                continue; // skip semi-infinite layers
            const auto slice_mat = result[i_slice].material();
            checkVolumeFractions(entry.second);
            result[i_slice].setMaterial(createAveragedMaterial(slice_mat, entry.second));
        }
    }
    ASSERT(result.size() == stack.size());
    return result;
}

//! Returns a SliceStack that refines the layer structure of sample,
//! taking into account the average SLD of layer material and particle decoration.
//!
//! For use in ProcessedSample constructor.

SliceStack slicify(const MultiLayer& sample, const SimulationOptions& options)
{
    SliceStack result;
    if (sample.numberOfLayers() == 0)
        return result;
    const bool use_slicing = options.useAvgMaterials() && sample.numberOfLayers() > 1;
    const auto layer_limits = SampleUtils::Preprocessor::particleRegions(sample, use_slicing);
    for (size_t i = 0; i < sample.numberOfLayers(); ++i) {
        const Layer* const layer = sample.layer(i);
        const size_t N = layer->numberOfSlices();
        if (N == 0)
            throw std::runtime_error("Cannot process layer with numberOfSlices=0");
        const ZLimits& slice_limits = layer_limits[i];
        double tl = layer->thickness();
        const Material* const material = layer->material();
        const LayerRoughness* roughness = SampleUtils::Multilayer::LayerTopRoughness(sample, i);
        if (roughness && roughness->getSigma() <= 0)
            roughness = nullptr;
        // if no slicing is needed, create single slice for the layer
        if (!slice_limits.isFinite() || N == 0) {
            if (i == sample.numberOfLayers() - 1)
                tl = 0.0;
            if (i == 0)
                result.addTopSlice(tl, *material);
            else
                result.addSlice(tl, *material, roughness);
            continue;
        }
        const double top = slice_limits.zTop();
        const double bottom = slice_limits.zBottom();
        // top layer
        if (i == 0) {
            if (top <= 0)
                throw std::runtime_error("ProcessedSample::ProcessedSample: "
                                         "top limit for top layer must be > 0.");
            result.addTopSlice(top, *material); // semiinfinite top slice
            result.addNSlices(N, top - bottom, *material);
            if (bottom > 0)
                result.addSlice(bottom, *material);
        }
        // middle or bottom layer
        else {
            if (top < 0) {
                result.addSlice(-top, *material, roughness);
                result.addNSlices(N, top - bottom, *material);
            } else {
                result.addNSlices(N, top - bottom, *material, roughness);
            }
            // middle layer
            if (i < sample.numberOfLayers() - 1 && bottom > -tl)
                result.addSlice(bottom + tl, *material);
            // bottom layer
            if (i == sample.numberOfLayers() - 1)
                result.addSlice(0.0, *material); // semiinfinite bottom slice
        }
    }
    return result;
}

//! Returns collection of all ProcessedLayout%s, for use in ProcessedSample constructor.

std::vector<ProcessedLayout> collectLayouts(const MultiLayer& sample, const SliceStack& slices,
                                            bool polarized)
{
    std::vector<ProcessedLayout> result;

    double z_ref = -slices.front().zBottom();

    for (size_t i = 0; i < sample.numberOfLayers(); ++i) {
        if (i > 1)
            z_ref -= sample.layer(i - 1)->thickness();
        for (const auto* layout : sample.layer(i)->layouts())
            result.emplace_back(ProcessedLayout(*layout, slices, z_ref, polarized));
    }
    return result;
}

} // namespace


//  ************************************************************************************************
//  class implementation
//  ************************************************************************************************

ProcessedSample ProcessedSample::make(const MultiLayer& sample, const SimulationOptions& options,
                                      bool forcePolarized)
{
    const bool polarized = forcePolarized || sample.isMagnetic() ||
        sample.externalField() != kvector_t();

    // slices1: accounting for roughness, but not yet for particles
    const SliceStack slices1 = slicify(sample, options).setBField(sample.externalField());

    std::vector<ProcessedLayout> layouts = collectLayouts(sample, slices1, polarized);

    const SliceStack refined_stack =
        options.useAvgMaterials() ? refineStack(slices1, layouts) : slices1;

    SliceStack reversed_stack;
    for (Slice slice : refined_stack) {
        slice.invertBField();
        reversed_stack.push_back(slice);
    }

    std::unique_ptr<const ISpecularStrategy> specular_strategy;
    if (polarized)
         specular_strategy =
            SampleUtils::SpecularStrategyBuilder::buildMagnetic(sample.roughnessModel());
    else
         specular_strategy =
            SampleUtils::SpecularStrategyBuilder::buildScalar(sample.roughnessModel());

    std::unique_ptr<const IFresnelMap> fresnel_map;
    if (polarized)
        fresnel_map = std::make_unique<const MatrixFresnelMap>(
            refined_stack,
            SampleUtils::SpecularStrategyBuilder::buildMagnetic(sample.roughnessModel()));
    else
        fresnel_map = std::make_unique<const ScalarFresnelMap>(
            refined_stack,
            SampleUtils::SpecularStrategyBuilder::buildScalar(sample.roughnessModel()));

    return ProcessedSample(sample, polarized, std::move(layouts), refined_stack, reversed_stack,
                           specular_strategy.release(), fresnel_map.release());
}

ProcessedSample::ProcessedSample(const MultiLayer& sample, bool polarized,
                                 std::vector<ProcessedLayout>&& layouts,
                                 const SliceStack& refined_stack,
                                 const SliceStack& reversed_stack,
                                 const ISpecularStrategy* specular_strategy,
                                 const IFresnelMap* fresnel_map)
    : m_sample(sample)
    , m_polarized(polarized)
    , m_layouts(std::move(layouts))
    , m_refined_stack(refined_stack)
    , m_reversed_stack(reversed_stack)
    , m_specular_strategy(specular_strategy)
    , m_fresnel_map(fresnel_map)
{
}

ProcessedSample::ProcessedSample(ProcessedSample&&)
    : ProcessedSample(m_sample, m_polarized, std::move(m_layouts), m_refined_stack,
                      m_reversed_stack, m_specular_strategy.release(), m_fresnel_map.release())
{
}

ProcessedSample::~ProcessedSample() = default;

size_t ProcessedSample::numberOfSlices() const
{
    return m_refined_stack.size();
}

const SliceStack& ProcessedSample::averageSlices() const
{
    return m_refined_stack;
}

const Slice& ProcessedSample::avgeSlice(size_t i) const
{
    return m_refined_stack.at(i);
}

const std::vector<ProcessedLayout>& ProcessedSample::layouts() const
{
    return m_layouts;
}

const ISpecularStrategy* ProcessedSample::getSpecularStrategy() const
{
    return m_specular_strategy.get();
}

const IFresnelMap* ProcessedSample::fresnelMap() const
{
    return m_fresnel_map.get();
}

double ProcessedSample::sliceTopZ(size_t i) const
{
    return m_refined_stack.at(i).zTop();
}

double ProcessedSample::sliceBottomZ(size_t i) const
{
    return m_refined_stack.at(i).zBottom();
}

bool ProcessedSample::containsMagneticMaterial() const
{
    return m_polarized;
}

bool ProcessedSample::polarizing() const
{
    return containsMagneticMaterial() || m_sample.externalField() != kvector_t{};
}

bool ProcessedSample::hasRoughness() const
{
    for (const auto& slice : m_refined_stack)
        if (slice.topRoughness())
            return true;
    return false;
}

Fluxes ProcessedSample::fluxesIn(const kvector_t& k) const
{
    return m_specular_strategy->Execute(m_refined_stack, k);
}

Fluxes ProcessedSample::fluxesOut(const kvector_t& k) const
{
    return m_specular_strategy->Execute(m_reversed_stack, -k);
}

double ProcessedSample::crossCorrSpectralFun(const kvector_t kvec, size_t j, size_t k) const
{
    const double xCorrLength = m_sample.crossCorrLength();
    if (xCorrLength <= 0.0)
        return 0.0;
    const double z_j = sliceBottomZ(j);
    const double z_k = sliceBottomZ(k);
    const LayerRoughness* rough_j = m_refined_stack.at(j + 1).topRoughness();
    const LayerRoughness* rough_k = m_refined_stack.at(k + 1).topRoughness();
    if (!rough_j || !rough_k)
        return 0.0;
    const double sigma_j = rough_j->getSigma();
    const double sigma_k = rough_k->getSigma();
    if (sigma_j <= 0 || sigma_k <= 0)
        return 0.0;
    return 0.5
           * ((sigma_k / sigma_j) * rough_j->getSpectralFun(kvec)
              + (sigma_j / sigma_k) * rough_k->getSpectralFun(kvec))
           * std::exp(-1 * std::abs(z_j - z_k) / xCorrLength);
}
